/* A quick & dirty test implementation of a fast software blur described
 * by Alex Evans in the article "Four Tricks for Fast Blurring in Software and
 * Hardware" at Gamasutra.
 * http://www.gamasutra.com/features/20010209/evans_02.htm
 *
 * Author: John Tsiombikas <nuclear@siggraph.org> - 24 Jan. 2005
 * This code is public domain, feel free to use it any way you like.
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include "timer.h"
#include "fps_counter.h"

/* screen resolution */
#define SCRX	800
#define SCRY	600

/* size of the texture used to retrieve the framebuffer */
#define TEX_XSZ	256
#define TEX_YSZ	256

enum {
	BLUR_BOTH,	/* blur in X and Y */
	BLUR_HORIZ,	/* blur in X */
	BLUR_VERT	/* blur in Y */
};

void update_display(void);
void key_handler(unsigned char k, int x, int y);
void material(float r, float g, float b, float spec);
void draw_scene(float t);
void draw_quad(void);
int blur_to_kernel_size(int amount);
void process_texture(unsigned long *buf, int x, int y, int scan_len);
void fast_blur(int dir, int amount, unsigned long *buf, int x, int y);
void update_title(void);

/* some text buffers for the titlebar state */
static const char *title_str = "Fast software blur test";
static char stats_buf[256] = "+/- to change blur amount";
static char fps_buf[256] = "fps: N/A";

unsigned int fbtex;	/* framebuffer aquisition texture */
int blur_amount = 0;
ntimer timer;
fps_counter fps;	/* for the fps counting system */

int main(int argc, char **argv) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(SCRX, SCRY);
	glutCreateWindow(title_str);
	update_title();
	
	glutDisplayFunc(update_display);
	glutIdleFunc(update_display);
	glutKeyboardFunc(key_handler);

	/* OpenGL state initialization */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CW);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0, 1.33333, 1.0, 1000.0);

	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClearDepth(1.0);

	/* prepare the texture for the fb copy */
	glGenTextures(1, &fbtex);
	glBindTexture(GL_TEXTURE_2D, fbtex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	{
		unsigned long *foo = malloc(TEX_XSZ * TEX_YSZ * 4);
		glTexImage2D(GL_TEXTURE_2D, 0, 4, TEX_XSZ, TEX_YSZ, 0, GL_RGBA, GL_UNSIGNED_BYTE, foo);
		free(foo);
	}

	timer_reset(&timer);
	timer_start(&timer);

	fps_start(&fps, 0, 1000);

	glutMainLoop();
	return 0;
}

void update_display(void) {
	static unsigned long buffer[TEX_XSZ * TEX_YSZ];
	unsigned long msec = timer_getmsec(&timer);
	float t = (float)msec / 1000.0f;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	if(blur_amount) glViewport(0, 0, TEX_XSZ, TEX_YSZ);
	draw_scene(t);
	if(blur_amount) glViewport(0, 0, SCRX, SCRY);

	if(blur_amount) {
		glBindTexture(GL_TEXTURE_2D, fbtex);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, TEX_XSZ, TEX_YSZ);

		glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
		process_texture(buffer, TEX_XSZ, TEX_YSZ, TEX_XSZ);
		glTexImage2D(GL_TEXTURE_2D, 0, 4, TEX_XSZ, TEX_YSZ, 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		draw_quad();
	}

	glutSwapBuffers();

	if(fps_frame_proc(&fps, timer_getmsec(&timer))) {
		sprintf(fps_buf, "fps: %f", fps_get_frame_rate(&fps));
		update_title();
	}
}

int blur_to_kernel_size(int amount) {
	amount *= 2;
	return amount % 2 ? amount : amount + 1;
}

void key_handler(unsigned char k, int x, int y) {
	switch(k) {
	case 27:
	case 'q':
	case 'Q':
		exit(0);

	case '=':
	case '+':
		blur_amount++;
		if(blur_amount >= TEX_XSZ / 2) blur_amount = TEX_XSZ / 2 - 1;
		if(blur_amount >= TEX_YSZ / 2) blur_amount = TEX_YSZ / 2 - 1;
		sprintf(stats_buf, "blur: %d, kernel size: %d", blur_amount, blur_to_kernel_size(blur_amount));
		update_title();
		break;

	case '-':
	case '_':
		blur_amount--;
		if(blur_amount < 0) blur_amount = 0;
		sprintf(stats_buf, "blur: %d, kernel size: %d", blur_amount, blur_to_kernel_size(blur_amount));
		update_title();
		break;

	default:
		break;
	}
}

void material(float r, float g, float b, float spec) {
	static float white[] = {1.0, 1.0, 1.0, 1.0};
	float col[4];
	col[0] = r;
	col[1] = g;
	col[2] = b;
	col[3] = 1.0;
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, col);
	glMaterialfv(GL_FRONT, GL_SPECULAR, white);
	glMaterialf(GL_FRONT, GL_SHININESS, spec);
}

#define SPH_USUB	23
#define SPH_VSUB	18
#define SPH_OFFSET	2.0
#define SPH_RAD		0.35

void draw_scene(float t) {
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glTranslatef(0, 0, -5);
	glRotatef(30 + sin(t * 2.0) * 20.0, 1.0, 0.0, 0.0);
	glRotatef(t * 100.0, 0.0, 1.0, 0.0);

	material(0.6, 0.5, 0.5, 70.0);
	glutSolidTeapot(1.0);

	glFrontFace(GL_CCW);
	
	glPushMatrix();
	glTranslatef(SPH_OFFSET, 0.0, 0.0);
	material(1.0, 0.3, 0.2, 50.0);
	glutSolidSphere(SPH_RAD, SPH_USUB, SPH_VSUB);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-SPH_OFFSET, 0.0, 0.0);
	material(0.3, 0.2, 1.0, 50.0);
	glutSolidSphere(SPH_RAD, SPH_USUB, SPH_VSUB);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0.0, 0.0, SPH_OFFSET);
	material(0.9, 0.8, 0.2, 50.0);
	glutSolidSphere(SPH_RAD, SPH_USUB, SPH_VSUB);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0.0, 0.0, -SPH_OFFSET);
	material(0.2, 0.9, 0.1, 50.0);
	glutSolidSphere(SPH_RAD, SPH_USUB, SPH_VSUB);
	glPopMatrix();

	glFrontFace(GL_CW);
}

void draw_quad(void) {
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	
	glBegin(GL_QUADS);
	glColor4f(1, 1, 1, 1);

	glTexCoord2f(0, 0);
	glVertex3f(-3.1, -2.3, -5.55);

	glTexCoord2f(0, 1);
	glVertex3f(-3.1, 2.3, -5.55);

	glTexCoord2f(1, 1);
	glVertex3f(3.1, 2.3, -5.55);

	glTexCoord2f(1, 0);
	glVertex3f(3.1, -2.3, -5.55);
	glEnd();
	
	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
}

void process_texture(unsigned long *buf, int x, int y, int scan_len) {
	int kernel_size = blur_to_kernel_size(blur_amount);
	fast_blur(0, kernel_size, buf, x, y);
}

/* some color packing/unpacking macros */
#define RED(p)		(((p) >> 16) & 0xff)
#define GREEN(p)	(((p) >> 8) & 0xff)
#define BLUE(p)		((p) & 0xff)
#define RGB(r, g, b)	(((r) & 0xff) << 16) | (((g) & 0xff) << 8) | ((b) & 0xff)

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) > (b) ? (a) : (b))

/* dir can be BLUR_BOTH (0), BLUR_HORIZ, BLUR_VERT
 * amount must be odd and less than min(x, y)
 */
void fast_blur(int dir, int amount, unsigned long *buf, int x, int y) {
	int i, j, half;
	unsigned long *dptr, *sptr, *tmp_buf;

	int blur_len = dir == BLUR_HORIZ ? x : y;
	int blur_times = dir == BLUR_HORIZ ? y : x;

	if(amount <= 1) return;

	dptr = buf;
	half = amount / 2;

	tmp_buf = malloc(blur_len * sizeof(unsigned long));

	for(i=0; i<blur_times; i++) {
		int ar = 0, ag = 0, ab = 0;
		int divisor = 0;

		if(dir == BLUR_HORIZ) {
			sptr = tmp_buf;
			memcpy(sptr, dptr, x * sizeof(unsigned long));
		} else {
			dptr = buf + i;
		
			sptr = tmp_buf;
			for(j=0; j<y; j++) {
				*sptr++ = *dptr;
				dptr += x;
			}
			dptr = buf + i;
			sptr = tmp_buf;
		}


		for(j=0; j<half; j++) {
			unsigned long pixel = tmp_buf[j];
			ar += RED(pixel);
			ag += GREEN(pixel);
			ab += BLUE(pixel);
			divisor++;
		}

		for(j=0; j<blur_len; j++) {
			int r, g, b;
			
			if(j > half) {
				unsigned long out = *(sptr - half - 1);
				ar -= RED(out);
				ag -= GREEN(out);
				ab -= BLUE(out);
				divisor--;
			}

			if(j < x - half) {
				unsigned long in = *(sptr + half);
				ar += RED(in);
				ag += GREEN(in);
				ab += BLUE(in);
				divisor++;
			}

			r = ar / divisor;
			g = ag / divisor;
			b = ab / divisor;

			r = MAX(MIN(r, 255), 0);
			g = MAX(MIN(g, 255), 0);
			b = MAX(MIN(b, 255), 0);
			
			*dptr = RGB(r, g, b);
			dptr += dir == BLUR_HORIZ ? 1 : x;
			sptr++;
		}

	}

	free(tmp_buf);

	if(dir == BLUR_BOTH) fast_blur(BLUR_HORIZ, amount, buf, x, y);
}


void update_title(void) {
	static char buf[512];

	sprintf(buf, "%s :: %s :: %s", title_str, stats_buf, fps_buf);
	glutSetWindowTitle(buf);
}
